# $Id: Geronimo.pm,v 1.53 2003/02/27 23:57:23 after Exp $
#
# Copyright (C) 2002 Nicolas Delon
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Geronimo;

use strict;
use Tk;
use Tk::NoteBook;
use IO::Select;
use Geronimo::Netsoul;
use Geronimo::Conf;
use Geronimo::Contact;
use Geronimo::Talk;
use Geronimo::Skin;

# NONACTIVE means an idle or lock state
sub	SELF_STATE_ACTIVE		{ 0 }
sub	SELF_STATE_AUTO_NONACTIVE	{ 1 }
sub	SELF_STATE_MANUAL_NONACTIVE	{ 2 }

sub	new
{
    my	$class = shift;
    my	$self = { };
    
    bless($self, $class);
    $self->{auto_idle_support} = defined(eval("require Idle") && Idle::init());
    $SIG{CHLD} = sub { wait };
    $self->{self_state} = &SELF_STATE_ACTIVE;
    $self->{do_r} = $self->{do_w} = 0;
    $self->{log_output} = undef;
    $self->{user_disconnect} = 0;
    $self->load_conf or return 0;
    $self->load_skin or return 0;
    $self->build_gui;
    $self->build_contact_list;
    # init_ns need the top_window which is build by build_gui
    $self->init_ns;
    return $self;
}

sub	get_contact_prev_frame
{
    my	$self = shift;
    my	$login = shift;
    my	$prev_frame;
    
    foreach ($self->{conf}->get_ordered_contact_list)
    {
	$_ eq $login and last;
	$prev_frame = $self->{contact_list}{$_}->get_last_frame;
    }
    return $prev_frame;
}

sub	manual_change_state
{
    my	$self = shift;
    my	$state = shift;
    
    $self->{self_state} = (($state eq &ST_ACTIVE) ? 
			   &SELF_STATE_ACTIVE : 
			   &SELF_STATE_MANUAL_NONACTIVE);
    $self->{ns}->set_state(-state_st => $state);
}

sub	auto_change_state
{
    my	$self = shift;
    my	$state = shift;
    
    $self->{self_state} = (($state eq &ST_ACTIVE) ?
			   &SELF_STATE_ACTIVE :
			   &SELF_STATE_AUTO_NONACTIVE);
    $self->{ns}->set_state(-state_st => $state);
}

sub	warn
{
    my	$self = shift;
    my	$msg = shift;
    my	$win;
    
    $win = new MainWindow(-title => '( Geronimo ) Careful !');
    $win->Label(
	-text => $msg,
	-justify => 'left',
    )->pack(-fill => 'both', -expand => 1);
    $win->Button(-text => 'OK', -command => sub { $win->destroy }
		 )->pack(-fill => 'both', -expand => 1);
}

# search for contact
# 1: nickname 
# 2: login
# returns $login if login is not a contact

sub	get_name
{
    my	$self = shift;
    my	$login = shift;
    
    return $self->{contact_list}->{$login}->get_name
	if (defined $self->{contact_list}->{$login});
    return $login;
}

# search for contact nickname
# returns nickname if login is a contact and has a nickname
# returns undef otherwise

sub	get_nickname
{
    my	$self = shift;
    my	$login = shift;
    
    return $self->{contact_list}->{$login}->get_nickname
	if (defined $self->{contact_list}->{$login});
    return undef;
}

sub	get_contact
{
    my	$self = shift;
    my	$login = shift;
    
    $self->{contact_list}->{$login};
}

sub	get_state
{
    my $self = shift;
    my $login = shift;
    my $id = shift;
    
    $self->{contact_list}->{$login}->get_state($id);
}

sub	load_conf
{
    my $self = shift;
    
    $self->{conf} = new Geronimo::Conf;
    $self->{conf}->load;
    $self->{name} = $self->{conf}->get_nickname || $self->{conf}->get_login;
}

sub    load_skin
{
    my $self = shift;
    
    $self->{skin} = new Geronimo::Skin(-skin => $self->{conf}->get_skin);
    $self->{skin}->load and return 1;
    CORE::warn("can't load skin file\n") and return 0;
}

sub	init_ns
{
    my $self = shift;

    $self->{ns} = new Geronimo::Netsoul(
	-login =>		$self->{conf}->get_login,
	-passwd =>		$self->{conf}->get_passwd,
	-ns_host =>		$self->{conf}->get_ns_host,
	-ns_port =>		$self->{conf}->get_ns_port,
	-local_addr =>		$self->{conf}->get_local_addr,
	-location =>		$self->{conf}->get_location,
	-data =>
	"geronimo-$Geronimo::Netsoul::VERSION [". $self->{conf}->get_data ."]",
	-state_st =>		&ST_ACTIVE,
	-event_handler =>	[\&event_handler, $self],
	-request_handler =>	[\&request_handler, $self],
	-log_handler =>		[\&log_handler, $self],
    );
    $self->log_handler("=== geronimo started\n");
    $self->initiate_ns_connect(1);
    $self->{top_window}->repeat(1 * 1000, [\&update_self_state, $self]);
    $self->{top_window}->repeat(1 * 100, [\&select_dirty_hack, $self]);
}

# Initiate the netsoul connection. Since this sub is also responsible for
# engaging the critical event_handler mecanism, we print some debug messages
sub	initiate_ns_connect
{
    my $self = shift;
    my $delay_if_necessary = shift;

    if ($delay_if_necessary and $self->{ns}->get_nb_connect_attempts > 0) {
	return $self->delay_ns_connect;
    }
    $self->log_handler("trying to connect ns object\n");
    if (!$self->{ns}->connect) {
	$self->log_handler("ns object connection failed\n");
	    return $self->delay_ns_connect;
    }
}

sub	delay_ns_connect
{
    my $self = shift;

    $self->{label_info_str} =
    "next connection in " . $self->{conf}->get_reconnect_delay . "s";
    $self->log_handler("next connection attempt in "
	. $self->{conf}->get_reconnect_delay . " secs\n");
    $self->{top_window}->after($self->{conf}->get_reconnect_delay*1000,
	[\&initiate_ns_connect, $self, 0]);
    return 0;
}

# The dirty hack in question (the code is borrowed from gerod, but with
# a zero timeout). This sub is repeatedly called forever, even if the
# connection is down. There is no need to check for the state of the
# connection, because if it is down, do_r = do_w = 0 and the select is
# a NOP.
sub	select_dirty_hack
{
    my $self = shift;
    my $sel = IO::Select->new($self->{ns}->get_socket_handle);
    my @res = IO::Select::select
    ($self->{do_r}?$sel:undef, $self->{do_w}?$sel:undef, undef, 0);
    if (3 == @res) {$self->{ns}->read_socket if (!$#{$res[0]});
		   $self->{ns}->write_socket if (!$#{$res[1]});}
}

# Event handler called by the Netsoul object.
sub	event_handler
{
    my $self = shift;
    my $ev = shift;
    my %infos = @_;
    my $contact;

    # not all events provide $infos{login}
    $contact = $self->{contact_list}{$infos{login}} if (defined $infos{login});
    if ($ev == &EV_CONN) {
	$self->log_handler("connection state has changed to '"
	    . Geronimo::Netsoul::str_of_conn($infos{conn_state}) . "'\n");
	$self->update_label_info_str;
	if ($infos{conn_state} == CONN_DOWN) {
	    if ($self->{user_disconnect}) {
		$self->log_handler("connection shut down by user\n");
	    } else {
		$self->log_handler("connection shut down abnormally\n");
		if (defined $self->{ns}->{last_errmsg}) {
		    $self->warn("The server has returned this error:\n\n"
			. "\t$self->{ns}->{last_errmsg}\t\n\n"
			. "Please take an appropriate action: check your "
			. "password, etc.");
		} else {
		    $self->disconnection;
		    $self->initiate_ns_connect(1);
		}
	    }
	} elsif ($infos{conn_state} == CONN_READY) {
	    $self->init_contact_list;
	    $self->{ns}->watch_log(
		-logins => [$self->{conf}->get_ordered_contact_list]);
	}
    } elsif ($ev == &EV_LOGIN) {
	my $event_login_remainder = sub {
	    my $contact = shift;
	    my $id = shift;
	    CORE::warn "'list_users :$id' has returned ". scalar @_ ." results"
	    if (@_ != 1);
	    $contact->update_from_list_users($id, $_[0]) if (@_ >= 1);
	};
	$contact->new_entry(\%infos);
	$self->{ns}->list_users(
	    -res_handler => [$event_login_remainder, $contact, $infos{id}],
	    -logins => [":".$infos{id}]);
    } elsif ($ev == &EV_LOGOUT) {
	$self->{talk}->disconnection($infos{id})
	if(defined $self->{talk} && $self->{talk}->have_talk_entry($infos{id}));
	$contact->delete_entry($infos{id}) if ($contact->has_entry($infos{id}));
    } elsif ($ev == &EV_MSG) {
	$self->recv_msg(\%infos);
    } elsif ($ev == &EV_STATE) {
	$contact->update_from_state_event($infos{id}, \%infos)
	if ($contact->has_entry($infos{id}));
    } elsif ($ev == &EV_TALK) {
	$self->disp_nstalk(\%infos);
    } elsif ($ev == &EV_TALKANS) {
	# not implemented
    } elsif ($ev == &EV_NEWMAIL) {
	$self->disp_mail(\%infos) if ($self->{conf}->get_mail_popup);
	print "\a" if ($self->{conf}->get_mail_beep);
    } elsif ($ev == &EV_PING) {
	$self->{ns}->send_ping;
    } else {
	$self->log_handler("unknown event $ev\n");
    }
}

#
# Request handler called by the Netsoul object.
#
# There are various problems and bugs in the Tk::fileevent mecanism,
# for a quick overview:
#   http://www.rosat.mpe-garching.mpg.de/mailing-lists/ptk/2000-12/
#   msg00060.html
# Geronimo suffers from this one:
#   http://www.rosat.mpe-garching.mpg.de/mailing-lists/makemaker/2002-01/
#   msg00008.html
# This is a perl bug that impacts on self-tied handles used by Tk. It
# seems there is no work-around for the moment (february 2003), that's
# why we are using a dirty hack: a handler repeatedly called that
# select()s the socket handle.
#
sub	request_handler
{
    my $self = shift;
    my $req = shift;

    #print("before request_handler: do_r[$self->{do_r}] do_w[$self->{do_w}]\n");

    # Here is how fileevent would be used if there was no perl bug
#    {
#	($req == &REQ_RD_START)
#	    and $self->{top_window}->fileevent($self->{ns}->get_socket_handle,
#	    'readable', [\&Geronimo::Netsoul::read_socket, $self->{ns}]),
#	$self->{do_r} = 1, last;
#	($req == &REQ_RD_STOP)
#	    and $self->{top_window}->fileevent($self->{ns}->get_socket_handle,
#	    'readable', undef), $self->{do_r} = 0, last;
#	($req == &REQ_WR_START)
#	    and $self->{top_window}->fileevent($self->{ns}->get_socket_handle,
#	    'writable', [\&Geronimo::Netsoul::write_socket, $self->{ns}]),
#	$self->{do_w} = 1, last;
#	($req == &REQ_WR_STOP)
#	    and $self->{top_window}->fileevent($self->{ns}->get_socket_handle,
#	    'writable', undef), $self->{do_w} = 0, last;
#	CORE::warn("** bug ** unknown request $req");
#    }

    # Here is the little part of "dirty hack" that just set some flags
    {
	($req == &REQ_RD_START)	and $self->{do_r} = 1, last;
	($req == &REQ_RD_STOP)	and $self->{do_r} = 0, last;
	($req == &REQ_WR_START)	and $self->{do_w} = 1, last;
	($req == &REQ_WR_STOP)	and $self->{do_w} = 0, last;
	$self->log_handler("** bug ** unknown request $req\n");
    }

    #print("after request_handler:  do_r[$self->{do_r}] do_w[$self->{do_w}]\n");
}

# Log handler called by the Netsoul object.
sub	log_handler
{
    my $self = shift;
    my ($msg) = @_;

    # return if silent
    return if (!$self->{conf}->get_output);

    # if log_handler() is called for the first time, init $self->{log_output}
    if (!defined $self->{log_output}) {
	if ($self->{conf}->get_logfile eq '-') {
	    $self->{log_output} = \*STDOUT;
	} else {
	    $_ = $self->{conf}->get_logfile;
	    s/^~/$ENV{HOME}/e;
	    if (!open($self->{log_output}, '>>', $_)) {
		print "can't open '$_' for writing: $!";
		$self->{log_output} = \*STDOUT;
	    }
	}
	$self->{log_output}->autoflush(1);
    }

    # log the msg; note that actually, there is no difference between
    # 'normal' and 'verbose' logging
    print { $self->{log_output} } "[" . localtime() . "] $msg";
}

sub	init_contact_list
{
    my $self = shift;
    
    $self->{ns}->list_users(
	-res_handler => [\&init_contact_list_remainder, $self],
	-logins => [$self->{conf}->get_ordered_contact_list]);
}

sub init_contact_list_remainder
{
    my $self = shift;
    my @list = @_;
    my %hash;
    my $login;

    foreach (@list)
    {
	$hash{$_->{login}} = [ ] if (not defined $hash{$_->{login}});
	push(@{ $hash{$_->{login}} }, $_);
    }
    foreach $login ($self->{conf}->get_ordered_contact_list)
    {
	next unless (defined $hash{$login});
	foreach (@{ $hash{$login} })
	{
	    # "simulation" of EV_LOGIN by calling new_entry
	    $self->{contact_list}{$login}->new_entry($_);
	    # "simulation" of EV_STATE by calling update_from_list_users
	    $self->{contact_list}{$login}->update_from_list_users($_->{id}, $_);
	}
    }
}

sub	disconnection
{
    my	$self = shift;
    
    foreach (keys %{ $self->{contact_list} })
    {
	$self->{contact_list}{$_}->disconnection;
    }
}

sub	init_gui_menu_bar_state
{
    my	$self = shift;
    my  $skin = $self->{skin};
    my	$menu;
    my	@menuitems;
    
    if (Geronimo::Netsoul::get_op_mode() == &MODE_EXTERNAL)
    {
	@menuitems =
	(
	    [ 'command' =>	'Switch to Active',
	    -command =>	sub { $self->manual_change_state(&ST_ACTIVE) }
	    ],
	    [ 'command' =>	'Switch to Idle',
	    -command =>	sub { $self->manual_change_state(&ST_IDLE) }
	    ],
	    [ 'command' =>	'Switch to Lock',
	    -command =>	sub { $self->manual_change_state(&ST_LOCK) }
	    ],
	    '-'
	);
    }
    push(@menuitems,
	[ 'command' =>	'Send Msg',
	-command =>	sub { $self->send_to_non_contact }
	],
	'-',
	[ 'command' =>	'Exit',
	-command =>	\&exit, # it will call the destroy handler
	]);
    $self->{menu_bar_state} = $self->{menu_bar}->Menubutton(
	-text =>		'Action',
	-borderwidth =>		1,
	-activeforeground =>	$skin->{menu_htextcolor},
	-activebackground =>	$skin->{menu_hovercolor},
	-background =>		$skin->{menu_bgcolor},
	-foreground =>		$skin->{menu_textcolor},
	-font =>		($^O =~ /win/i)
	? "$skin->{menu_font}" . " $skin->{menu_fontsize}"
	. " $skin->{menu_fontweight}"
	: "-*-$skin->{menu_font}-$skin->{menu_fontweight}-R-Normal-*-"
	. "$skin->{menu_fontsize}-*-*-*-*-*-*-*"
    );
    $menu = $self->{menu_bar_state}->Menu(
	-tearoff =>		0,
	-borderwidth =>		1,
	-activeborderwidth =>	1,
	-background =>		$skin->{menu_bgcolor},
	-foreground =>		$skin->{menu_textcolor},
	-activeforeground =>	$skin->{menu_htextcolor},
	-activebackground =>	$skin->{menu_hovercolor},
	-font =>		($^O =~ /win/i)
	? "$skin->{menu_font}" . " $skin->{menu_fontsize}" .
	" $skin->{menu_fontweight}"
	: "-*-$skin->{menu_font}-$skin->{menu_fontweight}-R-Normal-*-"
	. "$skin->{menu_fontsize}-*-*-*-*-*-*-*",
	-menuitems =>		\@menuitems
    );
    $self->{menu_bar_state}->configure(-menu => $menu);
    $self->{menu_bar_state}->pack(-side => 'left');
}

sub	init_gui_menu_bar
{
    my	$self = shift;
    my  $skin = $self->{skin};
    
    $self->{menu_bar} = $self->{top_window}->Frame(
	-relief =>	'groove',
	-background =>	$self->{skin}->{menu_bgcolor},
	-borderwidth =>	1
    );
    $self->{menu_bar}->pack(
	-side =>	'top',
	-expand =>	0,
	-fill =>	'x'
    );
    $self->init_gui_menu_bar_state;
}

sub	update_label_info_str
{
    my	$self = shift;

    $self->{label_info_str} =
    Geronimo::Netsoul::str_of_conn($self->{ns}->get_conn_state);
}

sub	build_gui
{
    my	$self = shift;
    my  $skin = $self->{skin};
    my	$resz = $self->{conf}->get_resizable ? 1 : 0;

    $self->{top_window} = new MainWindow;
    $self->{top_window}->resizable($resz, $resz);
    $self->{top_window}->title("Geronimo-$Geronimo::Netsoul::VERSION");
    # the bind("<Destroy>", ...) method will call the destroy handler
    # for each the widgets in the destroyed window while the OnDestroy
    # method will call the destroy handler only once
    #$self->{top_window}->bind("<Destroy>", sub { destroy_handler($self) });
    $self->{top_window}->OnDestroy(sub { destroy_handler($self) });
    $self->{label_info_str} = "geronimo started";
    $self->init_gui_menu_bar;
    $self->{frame_contact_list} = $self->{top_window}->Frame(
	-relief =>	'raised',
	-width =>	150);
    $self->{label_info} = $self->{top_window}->Label(
	-textvariable => \$self->{label_info_str},
	-anchor => 'center',
	-width => 15,
	-relief => 'sunken',
	-borderwidth => 1,	 
	-foreground => $skin->{infofg},
	-background => $skin->{infobg},
	-font => ($^O =~ /win/i)
	?  "$skin->{info_font}" . " $skin->{info_fontsize}" .
	" $skin->{info_fontweight}"
	: "-*-$skin->{info_font}-$skin->{info_fontweight}-R-Normal-*-" .
	"$skin->{info_fontsize}-*-*-*-*-*-*-*");
    $self->{label_info}->pack(-fill => 'x', -expand => 0, -side => 'bottom');
    $self->{frame_contact_list}->pack(
	-side =>	'top',
	-expand =>	1,
	-fill =>	'both');
}

sub	build_contact_list
{
    my	$self = shift;
    my	%hash;
    my	@list;
    my	$login;
    
    @list = $self->{conf}->get_ordered_contact_list;
    %hash = $self->{conf}->get_contact_list;
    foreach $login (@list)
    {
	$self->{contact_list}->{$login} = new Geronimo::Contact(
	    -login =>		$login, 
	    -nickname =>	$hash{$login}->{nickname},
	    -geronimo =>	$self);
    }
}

sub	mainloop
{
    my	$self = shift;
    
    MainLoop;
}

sub	disp_msg_reply
{
    my	$self = shift;
    my	$nsinfo = shift;
    my	$contact = $self->get_contact($nsinfo->{login});
    my	$win;
    
    if ($contact)
    {
	if ($contact->get_state($nsinfo->{id}) ne &ST_DISCONNECTED)
	{
	     $self->send_msg(-contact => $contact, -id => $nsinfo->{id});
	}
	else
	{
	    $self->warn($contact->get_name
	       	. " disconnects from $nsinfo->{location}");
	}
    }
    else
    {
	$self->send_msg(-id => $nsinfo->{id}, -login => $nsinfo->{login},
			-location => $nsinfo->{location});
    }
}

sub	disp_msg_talk_mode
{
    my	$self = shift;
    my	$nsinfo = shift;
    my	$contact = $self->get_contact($nsinfo->{login});
    
    if ($contact && $contact->get_state($nsinfo->{id}) eq &ST_DISCONNECTED)
    {
	$self->warn("$nsinfo->{login} disconnects from $nsinfo->{location}");
    }
    else
    {
	$self->{talk} = new Geronimo::Talk(-geronimo => $self)
	if (not defined $self->{talk});
	$self->{talk}->add_entry(
	    -id => $nsinfo->{id},
	    -login => $nsinfo->{login},
	    -nickname => $self->get_nickname($nsinfo->{login}),
	    -location => $nsinfo->{location},
	);
	$self->{talk}->recv_msg($nsinfo->{id}, $nsinfo->{msg});
    }
}

sub	disp_msg
{
    my	$self = shift;
    my	$nsinfo = shift;
    my	$name = $self->get_name($nsinfo->{login});
    my	$time = sprintf("%.2d:%.2d:%.2d", (localtime)[2,1,0]);
    my	$win;
    my	$text;
    
    $win = new MainWindow(-title => "New message from $name");
    $win->Label(-text => "$name at $time")->pack(-side => 'top');
    $text = $win->Scrolled("Text", -scrollbars => 'e', -wrap => 'word',
			   -width => 40, -height => 5, );
    $text->insert('end', $nsinfo->{msg}); 
    $text->configure(-state => 'disable');
    $text->pack(-expand => 1, -fill => 'both', -side => 'top');

    $win->Button(-text => 'close', -command => sub { $win->destroy }
		 )->pack(-fill => 'x', -expand => 1, -side => 'bottom');

    $win->Button(
	-text => 'reply',
	-width => 10,
	-command => sub { $win->destroy, $self->disp_msg_reply($nsinfo) }
    )->pack(-fill => 'x', -expand => 1, -side => 'left');

    $win->Button(
	-text => 'talk mode',
	-width => 10,
	-command => sub { $win->destroy, $self->disp_msg_talk_mode($nsinfo) }
    )->pack(-fill => 'x', -expand => 1, -side => 'left');

    print "\a" if ($self->{conf}->get_msg_beep);
}

sub	disp_mail
{
    my	$self = shift;
    my	$nsinfo = shift;
    my	$time = sprintf("%.2d:%.2d:%.2d", (localtime)[2,1,0]);
    my	$win;
    
    $win = new MainWindow(-title => '( Geronimo ) Mail');    
    $win->Label(-text => "At $time\nNew mail from $nsinfo->{from}"
		)->pack(-expand => 1, -fill => 'both');
    $win->Button(-text => 'close', -command => sub { $win->destroy }
		 )->pack(-expand => 1, -fill => 'both');
}

sub	disp_nstalk
{
    my	$self = shift;
    my	$nsinfo = shift;
    my	$time = sprintf("%.2d:%.2d:%.2d", (localtime)[2,1,0]);
    my	$name = $self->get_name($nsinfo->{login});
    my	$win;
    my	$text;
    
    # check if the login is in the kill list
    return $self->reply_to_killed_login($nsinfo->{login})
    if (defined $self->{conf}->{kill_list}->{$nsinfo->{login}});

    $text = "At $time\nNstalk request from $name\@$nsinfo->{location}";
    $text .= "\n\n$nsinfo->{talkmsg}";
    $win = new MainWindow(-title => '( Geronimo ) Nstalk');
    $win->Label(-text => $text)->pack(-expand => 1, -fill => 'both');
    $win->Button(-text => 'close', -command => sub { $win->destroy }
		 )->pack(-expand => 1, -fill => 'both');
    print "\a";
}

sub	recv_msg
{
    my	$self = shift;
    my	$nsinfo = shift;
    
    # check if the login is in the kill list
    return $self->reply_to_killed_login($nsinfo->{login})
    if (defined $self->{conf}->{kill_list}->{$nsinfo->{login}});

    if (defined $self->{talk} && $self->{talk}->have_talk_entry($nsinfo->{id}))
    {
	$self->{talk}->recv_msg($nsinfo->{id}, $nsinfo->{msg});
    }
    else
    {
	$self->disp_msg($nsinfo);
    }
}

sub	reply_to_killed_login
{
    my $self = shift;
    my $login = shift;
    my $msg = $self->{conf}->{kill_list}->{$login}->{msg};

    if ($msg eq '') {
	$self->log_handler("msg from killed login '$login', ignoring\n");
    } else {
	$self->log_handler("msg from killed login '$login', auto-replying\n");
	$self->{ns}->send_msg(-login => $login, -msg => $msg);
    }
}

sub send_to_non_contact
{
  my $self = shift;
  my $win;
  my $login_entry;
  my $send;
  my $win_ok;
  my $win_cancel;
  
  $win = new MainWindow(-title => "Send msg to which login ?");      
  $login_entry = $win->Entry(-width => 40);       
  
  $send = sub 
    {
      my $dest_login = $login_entry->get;
      my $win_sub;
      my $text_entry;
      my $msg;
      my $wsub_ok;
      my $wsub_cancel;
      my $really_send;
      
      $win->destroy;
      $win_sub = new MainWindow(-title => "Send msg to $dest_login"); 
      $text_entry = $win_sub->Entry(-width => 40);    
      
      $really_send = sub
	{
	  my $msg = $text_entry->get;
	  $win_sub->destroy;
	  $self->{ns}->send_msg(
	      -login => $dest_login, 
	      -msg => $msg);
	};
      
      $text_entry->bind('<Return>' => $really_send);
      $text_entry->pack(-side => 'top', -expand => 0, -fill => 'x');
      $text_entry->focus;
      $wsub_cancel = $win_sub->Button(-text => "Cancel", 
				      -command => sub { $win_sub->destroy })
	      ->pack(-side => 'right', -expand => 1, -fill => 'both');
      
      $wsub_ok = $win_sub->Button(-text => "Send",
				  -width => 10, -command => $really_send)
	      ->pack(-side => 'left', -expand => 1, -fill => 'both');
    };
  
  $login_entry->bind('<Return>' => $send);
  $login_entry->pack(-side => 'top', -expand => 0, -fill => 'x');
  $login_entry->focus;
  $win_ok = $win->Button(-text => 'Okay', -width => 10,
			 -command => $send)
	  ->pack(-side => 'left', -expand => 1, -fill => 'both');
  $win_cancel = $win->Button(-text => 'Cancel', -width => 10,
			     -command => sub { $win->destroy })
	  ->pack(-side => 'right', -expand => 1, -fill => 'both');
}

sub	send_msg
{
    my	$self = shift;
    my	%opt = @_;
    my	$contact = $opt{-contact};
    my	$id = $opt{-id};
    my	$login = $contact ? $contact->{login} : $opt{-login};
    my	$location = $contact ? $contact->entry($id)->{location}:$opt{-location};
    my	$name = $contact ? $contact->get_name : $login;
    my	$w_win;
    my	$w_entry;
    my	$w_ok;
    my	$w_cancel;
    my	$send;
    
    $w_win = new MainWindow(-title => "send_msg $name");
    $w_entry = $w_win->Entry(-width => 40);
    $send =  sub {
	my $msg = $w_entry->get;
	
        $w_win->destroy;
	if ($contact && $contact->get_state($id) eq &ST_DISCONNECTED)
	{
	    $self->warn($contact->get_name . " disconnects from $location");
	    return;
	}
        $self->{ns}->send_msg(-id => $id, -msg => $msg);
    };
    $w_entry->bind('<Return>' => $send);
    $w_ok = $w_win->Button(-text => "Ok",
			   -command => $send);
    $w_cancel = $w_win->Button(-text => 'Cancel',
			       -command => sub { $w_win->destroy });
    $w_entry->pack(-side => 'top', -expand => 0, -fill => 'x');
    $w_entry->focus;
    $w_ok->pack(-side => 'left', -expand => 1, -fill => 'both');
    $w_cancel->pack(-side => 'right', -expand => 1, -fill => 'both');
}

sub	update_self_state
{
    my	$self = shift;
    my	$idle_time;
    
    return if (!$self->{auto_idle_support});
    $self->{self_state} == &SELF_STATE_MANUAL_NONACTIVE and return;
    $idle_time = Idle::get_idle_time() / 1000;
    if ($idle_time > $self->{conf}->get_idle_delay)
    {
	$self->auto_change_state(&ST_IDLE) 
	    if ($self->{self_state} == &SELF_STATE_ACTIVE);
    }
    else
    {
	$self->auto_change_state(&ST_ACTIVE)
	    if ($self->{self_state} == &SELF_STATE_AUTO_NONACTIVE);
    }
}

#
# Exit mechanism of geronimo:
#
#   Action/Exit
#        V
#     exit()    close the window
#        +--------------+
#        V
#  destroy_handler($self)
#        |
#        +-(if conn state != CONN_DOWN)-V
#        |                       ns->disconnect
#     (else)                            V
#        |                      EV_CONN/CONN_DOWN
#        +------------------------------+
#        V
# log_handler("geronimo stopped")
#
sub	destroy_handler
{
    my	$self = shift;

    if ($self->{ns}->get_conn_state != &CONN_DOWN) {
	$self->{user_disconnect} = 1;
	$self->{ns}->disconnect;
    }
    $self->log_handler("=== geronimo stopped\n");
}

1;
