# $Id: Contact.pm,v 1.28 2003/02/27 23:57:24 after Exp $
#
# Copyright (C) 2002 Nicolas Delon
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Geronimo::Contact;

use strict;
use diagnostics;
use POSIX;
use Tk;
use Geronimo::Talk;
use Geronimo::Netsoul;

sub	TYPE_NORMAL { 0 }
sub	TYPE_TALK   { 1 }

#
# There is 1 Contact object associated to each login. If a login is
# authentified 2 times, the same object will be used to store the 2
# entities.
#
sub	new
{
    my	$class = shift;
    my	%opt = @_;
    my	$self = {};
    
    bless($self, $class);
    $self->{geronimo} = $opt{-geronimo} || die;
    $self->{skin} = $self->{geronimo}->{skin};
    $self->{state_xpm} = {
	&ST_ACTIVE => "$self->{skin}->{skindir}/active.xpm",
	&ST_IDLE => "$self->{skin}->{skindir}/idle.xpm",
	&ST_LOCK => "$self->{skin}->{skindir}/locked.xpm",
	&ST_CONNECTION => "$self->{skin}->{skindir}/connection.xpm",
	&ST_SERVER => "$self->{skin}->{skindir}/server.xpm",
	&ST_UNKNOWN => "$self->{skin}->{skindir}/unknown.xpm",
	&ST_DISCONNECTED => "$self->{skin}->{skindir}/disconnected.xpm",
	};
    $self->{state_color} = {
	&ST_ACTIVE => "$self->{skin}->{color_active}",
	&ST_IDLE => "$self->{skin}->{color_idle}",
	&ST_LOCK => "$self->{skin}->{color_locked}",
	&ST_CONNECTION => "$self->{skin}->{color_connection}",
	&ST_SERVER => "$self->{skin}->{color_server}",
	&ST_UNKNOWN => "$self->{skin}->{color_unknown}",
	&ST_DISCONNECTED => "$self->{skin}->{color_disconnected}",
	};
    $self->{id_colors} = {
	&ST_ACTIVE => "$self->{skin}->{id_color_active}",
	&ST_IDLE => "$self->{skin}->{id_color_idle}",
	&ST_LOCK => "$self->{skin}->{id_color_locked}",
	&ST_CONNECTION => "$self->{skin}->{id_color_connection}",
	&ST_SERVER => "$self->{skin}->{id_color_server}",
	&ST_UNKNOWN => "$self->{skin}->{id_color_unknown}",
	&ST_DISCONNECTED => "$self->{skin}->{id_color_disconnected}",
	};
    $self->{login} = $opt{-login} || die;
    $self->{nickname} = $opt{-nickname} || undef;
    $self->{type} = $opt{-type} || &TYPE_NORMAL;
    $self->{entry_list} = [ ];
    $self->{entry_hash} = { };
    $self->create_default_entry;
    return $self;
}

sub	has_entry
{
    my	$self = shift;
    my	$id = shift;
    
    defined $self->{entry_hash}{$id};
}

sub	MenuAttr
{
    my	$self = shift;
    my	$skin = $self->{skin};
    
    (-relief =>			'groove',
     -activeborderwidth =>	1,
     -tearoff =>		0,
     -foreground =>		$skin->{menu_textcolor},
     -background =>		$skin->{menu_bgcolor},
     -activebackground =>	$skin->{menu_hovercolor},
     -activeforeground =>	$skin->{menu_htextcolor},
     -font => ($^O =~ /win/i) ? 
	    "$skin->{menu_font}" . " $skin->{menu_fontsize}" . " $skin->{menu_fontweight}" : 
  	    "-*-$skin->{menu_font}-$skin->{menu_fontweight}-R-Normal-*-$skin->{menu_fontsize}-*-*-*-*-*-*-*"
   );
}

sub	MenuEntryAttr
{
    my	$self = shift;
    my	$skin = $self->{skin};
    
    (-foreground =>		$skin->{menu_textcolor},
     -background =>		$skin->{menu_bgcolor},
     -activebackground =>	$skin->{menu_hovercolor},
     -activeforeground =>	$skin->{menu_htextcolor},
     -font => ($^O =~ /win/i) ? 
	  "$skin->{menu_font}" . " $skin->{menu_fontsize}" . " $skin->{menu_fontweight}" : 
	  "-*-$skin->{menu_font}-$skin->{menu_fontweight}-R-Normal-*-$skin->{menu_fontsize}-*-*-*-*-*-*-*"
 );
}

sub	get_last_frame
{
    my	$self = shift;
    
    foreach (reverse @{ $self->{entry_list} })
    {
	$_->{widget}{packed} and return $_->{widget}{frame};
    }
    return undef;
}

sub	get_prev_frame
{
    my	$self = shift;
    
    $self->get_last_frame or $self->{geronimo}->get_contact_prev_frame($self->{login});
}

sub	get_name
{
    my	$self = shift;
    
    $self->{nickname} || $self->{login};
}

sub	get_nickname
{
    (shift)->{nickname};
}

sub	get_state
{
    my	$self = shift;
    my	$id = shift;
    
    defined $self->{entry_hash}->{$id}
    ? $self->{entry_hash}->{$id}->{state_st}
    : &ST_DISCONNECTED;
}

sub	update_pack
{
    my	$self = shift;
    my	$widget = shift;
    my	$state = defined $_[0] ? shift : &ST_UNKNOWN;
    
    if ($self->{geronimo}{conf}->get_show($state))
    {
	if (not $widget->{packed})
	{
	    my $prev = $self->get_prev_frame;
	    
	    $widget->{frame}->pack(-expand => 1, -fill => 'both',
				   -side => 'top',
				   (defined $prev ? (-after => $prev) : ()));
	    $widget->{packed} = 1;
	    if (not $self->{geronimo}{conf}->get_resizable) {
		$self->{geronimo}{top_window}->geometry('');
	    }
	    $self->{geronimo}{top_window}->idletasks;
	}
    }
    else
    {
	if ($widget->{packed})
	{
	    $widget->{frame}->packForget;
	    $widget->{packed} = 0;
	    if (not $self->{geronimo}{conf}->get_resizable) {
		$self->{geronimo}{top_window}->geometry('');
	    }
	    $self->{geronimo}{top_window}->idletasks;
	}
    }
    # Note on the above code:
    #   Calling idletasks() seems to be necessary to make the top window
    # resize correctly, even if I really don't see any reason why we should
    # call it. It appears that in very particular situations, when multiple
    # widgets are created and packed during a single user timeslice (ie. when
    # the Tk event loop is not entered), Tk fails to update some of its
    # widgets. The fact that this happens only in "very particular" situations
    # lets me thinks that this may be an obscure Tk bug.
    #   On another plan, the geometry() method is called to automatically
    # resize the top window when this one is not resizable by the user.
}

sub	tk_build_new_entry
{
    my	$self = shift;
    #my	$prev = shift;
    my	$name = $self->{nickname} || $self->{login};
    my	$widget = { packed => 0 };
    my  $skin = $self->{skin};
    my	$width;
    my  %label_attr = (-text => $name, 
		       -relief => $skin->{useicons} ? 'flat' : 'groove',
		       -width => $skin->{clwidth}, -height => $skin->{clheight},
                       -background => $skin->{cbgcolor},
                       -anchor => $skin->{ctextalign});
    
    $widget->{frame} = $self->{geronimo}{frame_contact_list}->Frame;
    $widget->{frame}->configure(-background => $skin->{cbgcolor});
    if (defined $skin->{cfont} || defined $skin->{cfontsize} || defined $skin->{cfontweight})
    {
	$label_attr{-font} = $widget->{frame}->fontCreate
	    ((defined $skin->{cfont} ? (-family => $skin->{cfont}) : ()),
	     (defined $skin->{cfontsize} ? (-size => $skin->{cfontsize}) : ()),
	     (defined $skin->{cfontweight} ? (-weight => $skin->{cfontweight}) : ()));
    }
    if (not $skin->{useicons})
    {
	$widget->{state} = $widget->{frame}->Label
	    (-background => $self->{state_color}{&ST_DISCONNECTED},
	     -width => $skin->{slwidth}, -height => $skin->{slheight},
	     -relief => 'groove'	   );
    } 
    else
    {
	my $image = $widget->{frame}->Pixmap
	    (-file => $self->{state_xpm}{&ST_DISCONNECTED});
	# Note: we don't specify -width and -height for the label containing
	# the image because we let the widget adapts its size to the image size
	$widget->{state} = $widget->{frame}->Label
	    (-background => $self->{state_color}{&ST_DISCONNECTED},
	     -image => $image,
	     -relief => 'flat');
    }
    $widget->{state}->pack(-side => 'left', -expand => 1, -fill => 'both');
    $widget->{identity} = $widget->{frame}->Label(%label_attr);
    $widget->{identity}->pack(-expand => 1, -fill => 'both');
    $widget->{identity}->bind('<Enter>' => sub { 
	$widget->{identity}->configure(-relief => 'ridge') } );
    $widget->{identity}->bind('<Leave>' => sub { 
	$widget->{identity}->configure(-relief => $skin->{useicons} ? 'flat' : 'groove') } );
    if ($^O !~ /win/i)
    {
	my $menu = $widget->{identity}->Menu
	    ($self->MenuAttr,
	     -menuitems => 
	     [ [ 'command' => 'mail',
		 $self->MenuEntryAttr,
		 -command => [ \&send_mail, $self ]
	       ]
	     ]);
	$widget->{identity}->bind('<Button-3>' => sub { 
	    $menu->Popup(-popover => 'cursor',
			 -popanchor => 'nw') });
    }
    $widget->{identity}->configure(-foreground => $self->{id_colors}{&ST_DISCONNECTED});
    $self->update_pack($widget, &ST_DISCONNECTED);
    return $widget;
}

sub	build_disp_info_field
{
    my	$topwidget = shift;
    my($key, $value) = @_;
    my	$frame;
    my	$entry;
    my  $skin = $topwidget->{skin};
    
    $key =~ s/^(.{15}).*$/$1/;
    $frame = $topwidget->Frame(-relief => 'groove', -border => 1
			       )->pack(-expand => 1, -fill => 'both');
    $frame->Label( -width => 15, -text => $key
		  )->pack(-side => 'left', -expand => 1, -fill => 'both');
    if (length $value <= 30)
    {
	($entry = $frame->Entry(-width => 30)
	 )->pack(-side => 'left', -expand => 1, -fill => 'both');
	$entry->insert('end' => $value);
	$entry->configure(-state => 'disabled');
    }
    else
    {
	($entry = $frame->Scrolled('Entry', -scrollbars => 's',
				   -width => 30)
	 )->pack(-side => 'left', -expand => 1, -fill => 'both');
	$entry->Subwidget('xscrollbar')->configure(-width => 10);
	$entry->Subwidget('entry')->insert('end' => $value);
	$entry->Subwidget('entry')->configure(-state => 'disabled');
     }
}

sub	disp_info
{
    my	$self = shift;
    my	$entry = shift;
    my	$win;
    my	$frame;
    
    $win = new MainWindow(-title => "( Geronimo ) Info " . $self->get_name);
    build_disp_info_field($win, 'Login' => $self->{login});
    build_disp_info_field($win, 'Group' => $entry->{group});
    build_disp_info_field($win, 'Id' => $entry->{id});
    build_disp_info_field($win, 'Location' => $entry->{location});
    build_disp_info_field($win, 'IP' => $entry->{ip});
    build_disp_info_field($win, 'Login time' => scalar localtime($entry->{login_time}));
    build_disp_info_field($win, 'State' => $entry->{state});
    build_disp_info_field($win, 'State time' => (defined $entry->{state_time}
	    ? scalar localtime($entry->{state_time})
	    : "<not available>"));
    build_disp_info_field($win, 'User data' => $entry->{data});
    $win->Button(-text => 'close', -command => sub { $win->destroy }
		 )->pack(-expand => 1, -fill => 'both');
}

sub	send_mail
{
    my	$self = shift;
    my	$mua = $self->{geronimo}->{conf}->get_mua;
    my	$term = $self->{geronimo}->{conf}->get_term;
    my	$cmdline = "$term -e $mua $self->{login}\@epita.fr";
    
    exec $cmdline or POSIX::exit(1) if (! fork());
}

sub	tk_apply_new_state
{
    my	$self = shift;
    my	$entry = shift;
    my	$widget = $entry->{widget};
    my	$skin = $self->{skin};
    my	$state = defined $entry->{state_st} ? $entry->{state_st} : &ST_UNKNOWN;
    my	@menuitems;

    if (not $self->{skin}->{useicons}) {
	$widget->{state}->configure(
	    -background => $self->{state_color}{$state});
    } else {
	my  $image = $widget->{frame}->Pixmap(
	    -file => $self->{state_xpm}{$state});
	$widget->{state}->configure(
	    -background => $self->{state_color}{$state},
	    -image => $image);
    }
    $widget->{identity}->configure(-foreground => $self->{id_colors}{$state});
    if (!defined $entry->{state_st} || $entry->{state_st} ne &ST_DISCONNECTED) {
	$widget->{state}->bind('<Enter>' => sub { 
		my $str = defined $entry->{state} ? $entry->{state} : "<n/a>";
		if (defined $entry->{state_time}) {
		    my $period = $self->{geronimo}->{ns}->get_server_time -
		    $entry->{state_time};
		    $self->{geronimo}->{label_info_str} = sprintf
		    ("$str (%s%.2dh%.2dm)",
			($period < 0 ? '-' : ''),
			abs($period) / 3600,
			abs($period) % 3600 / 60);
		} else {
		    $self->{geronimo}->{label_info_str} = "$str (<n/a>)";
		}
	    });
	$widget->{state}->bind(
	    '<Leave>' => sub { $self->{geronimo}->update_label_info_str; });
	$widget->{identity}->bind(
	    '<Enter>' => sub {
		my $location = $entry->{location};

		$widget->{identity}->configure(-relief => 'ridge');
		$location =~ s/(.{12}).{4,}/$1.../;
		$self->{geronimo}->{label_info_str} =
		"$location ($entry->{ip_infos})";
	    }
	);
	$widget->{identity}->bind(
	    '<Leave>' => sub { 
		$widget->{identity}->configure(
		    -relief => $skin->{useicons} ? 'flat' : 'groove');
		$self->{geronimo}->update_label_info_str;
	    }
	);
	# binding for doubleclick on a login in the contact list
	$widget->{identity}->bind(
	    '<Double-Button-1>' => sub {
		$self->{geronimo}->send_msg(
		    -contact => $self, -id => $entry->{id});
	    }
	);
	push(@menuitems,
	    [ 'command' => 'info',
	    $self->MenuEntryAttr,
	    -command => [ \&disp_info, $self, $entry ] ]);
	push(@menuitems,
	    [ 'command' => 'mail',
	    $self->MenuEntryAttr,
	    -command => [ \&send_mail, $self ] ]) if ($^O !~ /win/i);
	push(@menuitems,
	    [ 'command' => 'msg',
	    $self->MenuEntryAttr,
	    -command => sub { $self->{geronimo}->send_msg(
		    -contact => $self, -id => $entry->{id}); }]);
	push(@menuitems,
	    [ 'command' => 'talk',
	    $self->MenuEntryAttr,
	    -command => sub {
		$self->{geronimo}->{talk} = new Geronimo::Talk(
		    -geronimo => $self->{geronimo})
		if (not defined $self->{geronimo}->{talk});
		$self->{geronimo}->{talk}->add_entry(
		    -id => $entry->{id},
		    -login => $self->{login},
		    -nickname => $self->get_name,
		    -location => $entry->{location});
	    }]);
	my $menu = $widget->{identity}->Menu(
	    $self->MenuAttr,
	    -menuitems => \@menuitems);
	$widget->{identity}->bind(
	    '<Button-3>' => sub { $menu->Popup(
		    -popover => 'cursor',
		    -popanchor => 'nw'); });
    }
    else
    {
	if ($^O !~ /win/i)
	{
	    my $menu;
	    
	    $menu = $widget->{identity}->Menu
		($self->MenuAttr,
		 -menuitems => 
		 [ [ 'command' => 'mail',
		     $self->MenuEntryAttr,
		     -command => [ \&send_mail, $self ],
		   ]
		 ]);
	    $widget->{identity}->bind('<Button-3>' => sub {
		$menu->Popup(-popover => 'cursor',
			     -popanchor => 'nw') });
	}
	$widget->{state}->bind('<Enter>' => "");
	$widget->{state}->bind('<Leave>' => "");
	$widget->{identity}->bind('<Enter>' => sub { 
	    $widget->{identity}->configure(-relief => 'ridge') } );
	$widget->{identity}->bind('<Leave>' => sub { 
	    $widget->{identity}->configure(-relief => $skin->{useicons} ? 'flat' : 'groove') } );
    }
    $self->update_pack($widget, $entry->{state_st});
}

# one create one default entry (in the state disconnected) per Contact
sub	create_default_entry
{
    my	$self = shift;
    my	$entry;
    
    $entry = { };
    $entry->{state_st} = &ST_DISCONNECTED;
    $entry->{widget} = $self->tk_build_new_entry;
    push(@{ $self->{entry_list} }, $entry);
}

#
# When external code want to add 1 authentified entity, it calls this sub.
#
sub	new_entry
{
    my	$self = shift;
    my	$nsinfo = shift;
    my	$entry;
    
    if (not defined $self->{entry_list}[0]->{id}) {
	# if the login associated to $self has not yet been seen by this
	# module, then one create a new entry
	$entry = $self->{entry_list}[0];
    } else {
	# if the login associated to $self has already been seen by this
	# module, then one adds another entry to the existing one(s)
	$entry = { };
	$entry->{widget} = $self->tk_build_new_entry;
	push(@{ $self->{entry_list} }, $entry);
    }
    $self->{entry_hash}->{$nsinfo->{id}} = $entry;
    # Note: EV_LOGIN does not provide all of these infos. The missing ones
    # (state, login_time, state_time, data) will be obtained from
    # update_from_list_users.
    foreach (qw(id group state ip location ip_infos login_time state_time data))
    {
	$entry->{$_} = $nsinfo->{$_};
    }
    $self->tk_apply_new_state($entry);
}

sub	delete_entry
{
    my	$self = shift;
    my	$id = shift;
    my	$entry = $self->{entry_hash}->{$id};
    my	$widget = $entry->{widget};
    
    delete $self->{entry_hash}->{$id};
    if (@{ $self->{entry_list} } == 1)
    {
	foreach (keys %{ $entry })
	{
	    $_ eq "widget" and next;
	    delete $entry->{$_};
	}
	$entry->{state_st} = &ST_DISCONNECTED;
	$self->tk_apply_new_state($entry);
    }
    else
    {
	my $idx = 0;
	
	$widget->{frame}->destroy;

	if (not $self->{geronimo}{conf}->get_resizable) {
	    $self->{geronimo}{top_window}->geometry('');
	}

	foreach (@{ $self->{entry_list} })
	{
	    if ($id == $_->{id})
	    {
		(splice(@{ $self->{entry_list} }, $idx, 1), last);
	    }
	    $idx++;
	}
    }
}

sub	disconnection
{
    my	$self = shift;
    
    foreach (keys %{ $self->{entry_hash} })
    {
	$self->delete_entry($_);
    }
}

# a new EV_STATE updates infos about a user
sub	update_from_state_event
{
    my	$self = shift;
    my	$id = shift;
    my	$nsinfo = shift;
    my	$entry = $self->{entry_hash}->{$id};
    
    foreach (qw(state state_st state_time)) {
	$entry->{$_} = $nsinfo->{$_};
    }
    $self->tk_apply_new_state($entry);
}

# a list_users reply updates infos about a user
sub	update_from_list_users
{
    my  $self = shift;
    my  $id = shift;
    my  $nsinfo = shift;
    my  $entry = $self->{entry_hash}->{$id};

    foreach (qw(login_time idle_time data state state_st state_time)) {
	$entry->{$_} = $nsinfo->{$_};
    }
    $self->tk_apply_new_state($entry);
}

sub	entry
{
    my	$self = shift;
    my	$id = shift;
    
    $self->{entry_hash}->{$id};
}

1;
