# $Id: Skin.pm,v 1.10 2003/03/08 20:06:41 after Exp $
#
# Copyright (C) 2002 Nicolas Delon
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Geronimo::Skin;

use strict;
use Geronimo::Conf;

my $STRING_PATTERN = q/[^'"\s]+/;	# simple str pattern
my $SQSTRING_PATTERN = q/'[^']*'/;	# simple quoted str pattern
my $DQSTRING_PATTERN = q/"[^"]*"/;	# double quoted str pattern

my %ANCHOR = (left => 'w', right => 'e', center => 'center');

sub	new
{
    my	$class = shift;
    my	%opt = @_;
    my	$self = {};
    
    bless($self, $class);

    $self->{skindir} = "$ENV{HOME}/.geronimo/skins/$opt{-skin}";
    $self->{cbgcolor} = "snow2";
    $self->{cfont} = undef;
    $self->{cfontweight} = undef;
    $self->{cfontsize} = undef;
    $self->{ctextalign} = "center";
    # Note: clwidth overwrites the 150 pixels width specified in
    # lib/Geronimo.pm::build_gui()
    $self->{clwidth} = 20;
    $self->{clheight} = 1;
    $self->{slwidth} = 2;
    $self->{slheight} = 2;
    $self->{info_font} = ($^O =~ /win/i) ? 'verdana' : 'helvetica';
    $self->{info_fontsize} = "12";
    $self->{info_fontweight} = "medium";
    $self->{color_active} = "green";
    $self->{color_idle} = "yellow";
    $self->{color_locked} = "orange";
    $self->{color_connection} = "red";
    $self->{color_server} = "red";
    $self->{color_unknown} = "red";
    $self->{color_disconnected} = "red";
    $self->{menu_bgcolor} = "snow2";
    $self->{menu_hovercolor} = "snow2";
    $self->{menu_textcolor} = "black";
    $self->{menu_htextcolor} = "black";
    $self->{menu_font} = ($^O =~ /win/i) ? 'verdana' : 'helvetica';
    $self->{menu_fontsize} = "12";
    $self->{menu_fontweight} = "medium";
    $self->{id_color_active} = "black";
    $self->{id_color_idle} = "black";
    $self->{id_color_locked} = "black";
    $self->{id_color_server} = "black";
    $self->{id_color_connection} = "black";
    $self->{id_color_unknown} = "black";
    $self->{id_color_disconnected} = "black";
    $self->{useicons} = 0;
    $self->{infobg} = "snow2";
    $self->{infofg} = "black";
    $self->{name} = "default skin";
    $self->{talk_bgcolor} = 'white';
    $self->{talk_inputbgcolor} = 'white';
    $self->{talk_font} = ($^O =~ /win/i) ? 'verdana' : 'helvetica';
    $self->{talk_fontsize} = 12;
    $self->{talk_fontweight} = 'medium';
    $self->{talk_fontweight2} = 'bold';
    $self->{talk_inputfontweight} = 'medium';
    $self->{talk_fontcolor} = 'black';
    $self->{talk_fontcolor2} = 'blue';
    $self->{talk_inputfontcolor} = 'black';
    $self->{talk_disconnectcolor} = 'red';
    return $self;
}

sub	print
{
    my	$self = shift;
    
    print "skin: ", $self->{skindir} || "<empty>", "\n"; 
    print "--skin: cfgcolor: [", $self->{cfgcolor} || "<empty>", "]\n";
    print "--skin: cbgcolor: [", $self->{cbgcolor} || "<empty>", "]\n";
    print "--skin: cfont: [", $self->{cfont} || "<empty>", "]\n";
    print "--skin: cfontweight: [", $self->{cfontweight} || "<empty>", "]\n";
    print "--skin: cfontsize: [", $self->{cfontsize} || "<empty>", "]\n";
    print "--skin: ctextalign: [", $self->{ctextalign} || "<empty>", "]\n";
    print "--skin: ssize: [", $self->{ssize} || "<empty>", "]\n";
    print "--skin: sbgcolor: [", $self->{sbgcolor} || "<empty>", "]\n";
    print "--skin: state colors: [", $self->{color_active}, "/", 
    		$self->{color_disconnected}, "/", $self->{color_idle}, "/",
    		$self->{color_locked}, "/", $self->{color_server}, "/",
		$self->{color_connection}, "/", $self->{color_unknown}, "]\n";
    print "--skin: menucolors: [", $self->{menu_bgcolor}, "/", 
    				   $self->{menu_hovercolor}, "]\n";
    print "--skin: info bgcolor: [", $self->{infobg}, "]\n";
    print "--skin: useicons: [", $self->{useicons} || "<empty>", "]\n";
    print "--skin: name: [", $self->{name} || "<empty>", "]\n";
}



sub	load
{
    my	$self = shift;
    my	%skin_parts = 	
	('contact.color' =>			[ 'cbgcolor', 'str', 'contact.color <color>' ],
	 'contact.font.family' =>		[ 'cfont', 'str', 'contact.font.family <font name>' ],
	 'contact.font.weight' =>		[ 'cfontweight', 'str', 'contact.font.weight <medium|bold>' ],
	 'contact.font.size' =>			[ 'cfontsize', 'int', 'contact.font.size <int value>' ],
	 'contact.text_alignment' =>		\&set_contactAlignment,
	 'contact.label.height' =>		[ 'clheight', 'int', 'contact.label.height <int value>' ],
	 'contact.label.width' =>		[ 'clwidth', 'int', 'contact.label.width <int value>' ],
	 'state.color.background' =>		[ 'sbgcolor', 'str', 'state.color.background <color>' ],
	 'state.useicons' =>			[ 'useicons', 'bool', 'state.useicon <yes|no>' ],
	 'state.color.idle'	=>		[ 'color_idle', 'str', 'state.color.idle <color>' ],
	 'state.color.active' =>		[ 'color_active', 'str', 'state.color.active <color>' ],
	 'state.color.disconnected'=>		[ 'color_disconnected', 'str', 'state.color.disconnected <color>' ],
	 'state.color.locked'	=>		[ 'color_locked', 'str', 'state.color.locked <color>' ],
	 'state.color.server'	=>		[ 'color_server', 'str', 'state.color.server <color>' ],
	 'state.color.connection'	=>		[ 'color_connection', 'str', 'state.color.connection <color>' ],
	 'state.color.unknown'	=>		[ 'color_unknown', 'str', 'state.color.unknown <color>' ],
	 'id.color.active' =>			[ 'id_color_active', 'str', 'id.color.active <color>' ],
	 'id.color.idle' =>			[ 'id_color_idle', 'str', 'id.color.idle <color>' ],
	 'id.color.locked' =>			[ 'id_color_locked', 'str', 'id.color.locked <color>' ],
	 'id.color.disconnected' =>		[ 'id_color_disconnected', 'str', 'id.color.disconnected <color>' ],
	 'id.color.server' =>		[ 'id_color_server', 'str', 'id.color.server <color>' ],
	 'id.color.connection' =>		[ 'id_color_connection', 'str', 'id.color.connection <color>' ],
	 'id.color.unknown' =>			[ 'id_color_unknown', 'str', 'id.color.unknown <color>' ],
	 'state.label.width' =>			[ 'slwidth', 'int', 'state.label.width <int value>' ],
	 'state.label.height' =>		[ 'slheight','int', 'state.label.height <int value>' ],
	 'menu.color.background' =>		[ 'menu_bgcolor', 'str', 'menu.color.background <color>' ],
	 'menu.color.hover' =>			[ 'menu_hovercolor', 'str', 'menu.color.hover <color>' ],
	 'menu.color.text' =>			[ 'menu_textcolor', 'str', 'menu.color.text <color>' ],
	 'menu.color.hovertext' =>		[ 'menu_htextcolor', 'str', 'menu.color.hovertext <color>' ],
	 'menu.font' =>				[ 'menu_font', 'str', 'menu.font <font name>' ],
	 'menu.font.size' =>			[ 'menu_fontsize', 'int', 'menu.font.size <int value>'],
	 'menu.font.weight' =>			[ 'menu_fontweight', 'str', 'menu.font.weight <medium|bold>' ],
	 'info.color' =>			[ 'infobg', 'str', 'info.color <color>' ],
	 'info.text' =>				[ 'infofg', 'str', 'info.text <color>' ],
	 'info.font' =>				[ 'info_font', 'str', 'info.font <font name>' ],
	 'info.font.size' =>			[ 'info_fontsize', 'int', 'info.font.size <int value>'],
	 'info.font.weight' =>			[ 'info_fontweight', 'str', 'info.font.weight <medium|bold>' ],
	 'skin.name' =>				[ 'name', 'str', 'skin.name <string>' ],
	 'talk.bgcolor' =>			[ 'talk_bgcolor', 'str', 'talk.bgcolor <color>' ],
	 'talk.input.bgcolor' =>		[ 'talk_inputbgcolor', 'str', 'talk.input.bgcolor <color>' ],
	 'talk.font' =>				[ 'talk_font', 'str', 'talk.font <font name>' ],
	 'talk.font.size' =>			[ 'talk_fontsize', 'int', 'talk.font.size <int value>'],
	 'talk.font.weight.self' =>		[ 'talk_fontweight', 'str', 'talk.font.weight.self <medium|bold>' ],
	 'talk.font.weight.buddy' =>		[ 'talk_fontweight2', 'str', 'talk.font.weight.buddy <medium|bold>' ],
	 'talk.font.weight.input' =>		[ 'talk_inputfontweight', 'str', 'talk.font.weight.input <medium|bold>' ], 
	 'talk.font.color.self' =>		[ 'talk_fontcolor', 'str', 'talk.font.color.self <color>' ],
	 'talk.font.color.buddy' =>		[ 'talk_fontcolor2', 'str', 'talk.font.color.buddy <color>' ],
	 'talk.font.color.input' =>		[ 'talk_inputfontcolor', 'str', 'talk.font.color.input <color>' ],
	 'talk.font.color.disconnect' =>	[ 'talk_disconnectcolor', 'str', 'talk.font.color.disconnect <color>' ]
	 );
    
    my	%hash;
    my	@list;
    my	$part;
    my	$retval;
    my	$lineno = 1;
    
    if (-d $self->{skindir})
    {
	if (!open(FILE, "$self->{skindir}/skin")) {
	    warn("can't open '$self->{skindir}/skin': $!\n");
	    return 0;
	}
	while (<FILE>)
	{  
	    chomp;
	    next if (/^\s*$/ or /^\s*\#/);
	    @list = /\s*($STRING_PATTERN|$SQSTRING_PATTERN|$DQSTRING_PATTERN)\s*/g;
	    s/^(\'|\")(.+)\1$/$2/ foreach (@list);
	    $part = shift @list;
	    defined $skin_parts{$part} or (warn "$part not recognized\n", next);
	    $retval = ((ref $skin_parts{$part} eq 'ARRAY')  ?
		       $self->set_value($lineno, $part, @{$skin_parts{$part}}, @list) :
		       &{$skin_parts{$part}}($self, $lineno, @list));
	    $lineno++;
	}
	return 1;
    }
    else
    {
	warn("'$self->{skindir}' is not a directory\n");
	return 0;
    }
}

sub set_talkFontSize 
{
    my	$self = shift;
    my	$lineno = shift;
    my	$value = shift;
    
    if ($value !~ /^\d+$/)
    {
	warn "Syntax error line $lineno, syntax: talk.font.size <int value>\n";
	return 0;
    }
    $self->{talk_fontsize} = int ($value + (($^O =~ /win/i) ? 0 : ($value / 2)));
    return 1;
}

sub set_contactAlignment
{
    my	$self = shift;
    my	$lineno = shift;
    my	$value = shift;
    
    if (not defined $ANCHOR{$value})
    {
	warn "Syntax error line $lineno, syntax: contact.text.alignment <left|right|center>\n";
	return 0; 
    }
    $self->{ctextalign} = $ANCHOR{$value};
    return 1;
}

sub	set_value
{
    my	$self = shift;
    my	$lineno = shift;
    my	$field = shift;
    my	$skinpart = shift;
    my	$value_type = shift;
    my	$syntax = shift;
    my	$value = shift || (warn "Syntax error line $lineno, syntax: $syntax\n", return 0);
    
    if ($value_type eq 'bool')
    {
	if ($value =~ /^yes|no|on|off|true|false|enable|disable$/i) {
	    $self->{$skinpart} = $value =~ /yes|on|true|enable/i;
	}
	else {
	    warn "Syntax error line $lineno, syntax: $syntax (boolean type)\n";
	    return 0;
	}
    }
    elsif ($value_type eq 'int')
    {
	if ($value =~ /^\d+$/) {
	    $self->{$skinpart} = int $value;
	}
	else {
	    warn "Syntax error line $lineno, syntax: $syntax (integer type)\n";
	    return 0;
	}
    }
    else { # string type 
	$self->{$skinpart} = $value;
    }
    return 1;
}

1;
