# $Id: Talk.pm,v 1.2 2003/02/08 20:42:49 after Exp $
#
# Copyright (C) 2002 Nicolas Delon
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Geronimo::Talk;

use strict;
use Tk;

sub	TAG_SEND		{ "tag_send" }
sub	TAG_RECV		{ "tag_recv" }
sub	TAG_DISCONNECTION	{ "tag_disconnection" }

sub	STATE_CONNECTED		{ 0 }
sub	STATE_DISCONNECTED	{ 1 }

sub	new
{
    my	$class = shift;
    my	%opt = @_;
    my	$self = { };
    my  $skin = { };
    
    bless($self, $class);
    $self->{geronimo} = $opt{-geronimo} || die;
    $self->{skin} = $self->{geronimo}->{skin};
    $skin = $self->{skin};
    $self->{title} = $opt{-title} || "( Geronimo ) Talk";
    $self->{win} = new MainWindow(-title => $self->{title});
    $self->{close} = $self->{win}->Button
	(-text => 'Close Current Page',
	 -command => sub {
	     ((keys %{ $self->{entry} }) == 1 ?
	      $self->{win}->destroy :
	      $self->close_current_page);
	 },
         -font => ($^O =~ /win/i) ? 
       "$skin->{talk_font}" . " $skin->{talk_fontsize}" . " $skin->{talk_fontweight}" :  
       "-*-$skin->{talk_font}-$skin->{talk_fontweight}-R-Normal-*-$skin->{talk_fontsize}-*-*-*-*-*-*-*"

       );
    $self->{notebook} = $self->{win}->NoteBook;
    $self->{close}->pack(-fill => 'x', -expand => 0, -side => 'bottom');
    $self->{notebook}->pack(-fill => 'both', -expand => 1, -side => 'top');
    $self->{notebook}->configure(
      -font => ($^O =~ /win/i) ? 
       "$skin->{talk_font}" . " $skin->{talk_fontsize}" . " $skin->{talk_fontweight}" :  
       "-*-$skin->{talk_font}-$skin->{talk_fontweight}-R-Normal-*-$skin->{talk_fontsize}-*-*-*-*-*-*-*"
  );
    $self->{win}->bind('<Destroy>' => sub { 
	undef $self->{geronimo}->{talk} if ($_[0] == $self->{win}) });
    $self->{entry} = { };
    return $self;
}

sub	have_talk_entry
{
    my	$self = shift;
    my	$id = shift;
    
    defined $self->{entry}{$id} && $self->{entry}{$id}{state} == &STATE_CONNECTED;
}

sub	add_text
{
    my	$self = shift;
    my	$widget = shift;
    my	$text = shift;
    my	$tag = shift;
    
    $widget->configure(-state => 'normal');
    $widget->insert('end' => $text, $tag);
    $widget->configure(-state => 'disable');
    $widget->yviewMoveto(1);
}

sub	add_entry
{
    my	$self = shift;
    my	%arg = @_;
    my	$id = $arg{-id};
    my	$location = $arg{-location};
    my	$login = $arg{-login};
    my	$nickname = $arg{-nickname};
    my	$msg = $arg{-msg};
    my	$name = $nickname || $login;
    my	$entry = { };
    my  $skin = $self->{skin};
    
    $self->close_page($id) if (defined $self->{entry}->{$id});
    $self->{entry}->{$id} = $entry;
    $entry->{login} = $login;
    $entry->{nickname} = $nickname;
    $entry->{name} = $name;
    $entry->{location} = $location;
    $entry->{state} = &STATE_CONNECTED;
    $entry->{page} = $self->{notebook}->add($id, -label => "$name\@$location");
    $entry->{text} = $entry->{page}->Scrolled('Text',
					      -state => 'disabled',
					      -wrap => 'word',
					      -scrollbars => 'e',
					      -background => $skin->{talk_bgcolor},
					      -foreground => $skin->{talk_fontcolor}
					    );

    $entry->{entry} = $entry->{page}->Entry(-background => $skin->{talk_inputbgcolor},
    					    -foreground => $skin->{talk_inputfontcolor},
				 	    -font => ($^O =~ /win/i) ? 
					       "$skin->{talk_font}" . " $skin->{talk_fontsize}" . " $skin->{talk_fontweight}" :  
	 			    	       "-*-$skin->{talk_font}-$skin->{talk_fontweight}-R-Normal-*-$skin->{talk_fontsize}-*-*-*-*-*-*-*"
					   );
    $entry->{text}->tagConfigure
	(&TAG_SEND, -foreground => $skin->{talk_fontcolor},
	 -font => ($^O =~ /win/i) ? 
	 			    "$skin->{talk_font}" . " $skin->{talk_fontsize}" . " $skin->{talk_fontweight}" : 
	 			    "-*-$skin->{talk_font}-$skin->{talk_fontweight}-R-Normal-*-$skin->{talk_fontsize}-*-*-*-*-*-*-*");

    $entry->{text}->tagConfigure
	(&TAG_RECV, -foreground => $skin->{talk_fontcolor2}, 
	 -font => ($^O =~ /win/i) ? 
	 			    "$skin->{talk_font}" . " $skin->{talk_fontsize}" . " $skin->{talk_fontweight2}" :  
	 			    "-*-$skin->{talk_font}-$skin->{talk_fontweight2}-R-Normal-*-$skin->{talk_fontsize}-*-*-*-*-*-*-*");

    $entry->{text}->tagConfigure
	(&TAG_DISCONNECTION, -foreground => $skin->{talk_disconnectcolor},
	 -font => ($^O =~ /win/i) ? 
	 		"$skin->{talk_font}" . " $skin->{talk_fontsize}" . " $skin->{talk_fontweight}" : 
	 		"-*-$skin->{talk_font}-Bold-R-Normal-*-$skin->{talk_fontsize}-*-*-*-*-*-*-*");

    $entry->{entry}->bind('<Return>', sub { 
	my $msg = $entry->{entry}->get();
	
	$self->send_msg($id, $msg);
	$entry->{entry}->delete(0, 'end') });
    $entry->{entry}->bind('<Expose>' => sub { 
	$self->{notebook}->pageconfigure($id, -label => "$name\@$entry->{location}");
	$entry->{entry}->focus } );
    $entry->{entry}->pack(-side => 'bottom', -fill => 'x', -expand => 0);
    $entry->{text}->pack(-fill => 'both', -expand => 1);
    $self->{notebook}->raise($id);
    $entry->{entry}->focus;
    $self->recv_msg($id, $msg) if (defined $msg);
}

sub	close_page
{
    my	$self = shift;
    my	$id = shift;
    
    delete $self->{entry}->{$id};
    $self->{notebook}->delete($id);
}

sub	close_current_page
{
    my	$self = shift;
    my	$id;
    
    $id = $self->{notebook}->raised or return;
    delete $self->{entry}->{$id};
    $self->{notebook}->delete($id);
}

sub	disconnection
{
    my	$self = shift;
    my	$id = shift;
    my	$entry = $self->{entry}->{$id};
    my	$time = sprintf("%.2d:%.2d:%.2d", (localtime)[2,1,0]);
    
    $self->add_text($entry->{text} => 
		    "*** $entry->{name}\@$entry->{location} has disconnected at $time ***\n",
		    &TAG_DISCONNECTION);
    $entry->{entry}->configure(-state => 'disable');
    $entry->{state} = &STATE_DISCONNECTED;
}

sub	add_msg
{
    my	$self = shift;
    my	$id = shift;
    my	$name = shift;
    my	$msg = shift;
    my	$tag = shift;
    my	$entry = $self->{entry}->{$id};
    my	$time = sprintf("%.2d:%.2d:%.2d", (localtime)[2,1,0]);
    
    $self->add_text($entry->{text} => "$name $time> $msg", $tag);
}

sub	send_msg
{
    my	$self = shift;
    my	$id = shift;
    my	$msg = shift;
    
    $self->{geronimo}->{ns}->send_msg(-id => $id, -msg => $msg);
    $msg =~ s[\\n][\n]g;
    $self->add_msg($id, $self->{geronimo}->{name}, "$msg\n", &TAG_SEND);
}

sub	recv_msg
{
    my	$self = shift;
    my	$id = shift;
    my	$msg = shift;
    my	$entry = $self->{entry}->{$id};
    my	$name = $entry->{name};
    
    $msg =~ /\n$/ or $msg .= "\n";
    $self->add_msg($id, $name, $msg, &TAG_RECV);
    $self->{notebook}->pageconfigure($id, -label => "$name\@$entry->{location} *")
	if ($self->{notebook}->raised != $id);
}

1;
